// SPDX-FileCopyrightText: 2021 GNOME Foundation
// SPDX-License-Identifier: LGPL-2.1-or-later

// A map between namespaces and base URLs for their online documentation
baseURLs = [
    [ 'GObject', 'https://docs.gtk.org/gobject/' ],
]
