/*
 * Copyright (C) 2017-2020 Red Hat
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2 of the
 * License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, see <http://www.gnu.org/licenses/>.
 */

#pragma once

#include <xf86drm.h>
#include <xf86drmMode.h>

#include "backends/meta-crtc-mode.h"
#include "backends/native/meta-kms-types.h"

#define META_TYPE_CRTC_MODE_KMS (meta_crtc_mode_kms_get_type ())
G_DECLARE_FINAL_TYPE (MetaCrtcModeKms, meta_crtc_mode_kms,
                      META, CRTC_MODE_KMS,
                      MetaCrtcMode)

MetaKmsMode * meta_crtc_mode_kms_get_kms_mode (MetaCrtcModeKms *mode_kms);

MetaCrtcModeKms * meta_crtc_mode_kms_new (MetaKmsMode             *kms_mode,
                                          MetaCrtcRefreshRateMode  refresh_rate_mode,
                                          uint64_t                 id);
